import glob 
import os
import os.path as osp 
from shutil import copyfile
import json 
from tqdm import tqdm
from mlearning.datasets.converters.polygon2dota import polygon2dota_by_rotate

def convert_labelme2dota(input_dir, output_dir, copy_image=True, image_ext='bmp'):
    if not osp.exists(output_dir):
        os.mkdir(output_dir)
        
    folders = [folder.split("/")[-1] for folder in glob.glob(osp.join(input_dir, "**")) if not osp.isfile(folder)]

    for folder in folders:
        _output_label_dir = osp.join(output_dir, folder, 'labelTxt')
        if not osp.exists(_output_label_dir):
            os.makedirs(_output_label_dir)

        if copy_image:
            _output_image_dir = osp.join(output_dir, folder, 'images')
            if not osp.exists(_output_image_dir):
                os.makedirs(_output_image_dir)

        json_files = glob.glob(osp.join(input_dir, folder, '*.json'))
        print(f"There are {len(json_files)} json files")

        for json_file in tqdm(json_files, desc=folder):
            
            assert osp.exists(json_file), ValueError(f"There is no such json-file: {json_file}")
            
            filename = osp.split(osp.splitext(json_file)[0])[-1]
            
            if copy_image:
                img_file = osp.splitext(json_file)[0] + f'.{image_ext}'
                assert osp.exists(json_file), ValueError(f"There is no such json-file: {json_file}")

                copyfile(img_file, osp.join(_output_image_dir, filename + f'.{image_ext}'))

            
            txt = open(osp.join(_output_label_dir, filename + '.txt'), 'w')
            with open(json_file, 'r') as jf:
                anns = json.load(jf)['shapes']
                
            if len(anns) != 0:
                for ann in anns:
                    points, _ = polygon2dota_by_rotate(ann['points'])
                    data = ''
                    for point in points:
                        data += f'{str(int(point[0]))} {str(int(point[1]))} '
                    
                    data += f'{ann["label"]} 0'
                    txt.write(data)
                    txt.write("\n")
            
        txt.close()

            



from pydantic import Field
from typing import Any, Literal

from aivcommon.configs.utils.helpers import string_to_list_of_type
from aivcommon.configs.src.bases import BaseConfig



class TrainConfig(BaseConfig):
    # required
    backbone: Literal['n', 's', 'm', 'l', 'x'] = Field(frozen=True)
    epochs: int = Field(frozen=True)
    batch: int = Field(frozen=True)
    imgsz: int = Field(frozen=True)
    device: str = Field('0', frozen=True)
    
    #
    conf: float = Field(0.25, frozen=True)
    iou: float = Field(0.7, frozen=True)

    # 
    lrf: float = Field(0.001, frozen=True)
    label_smoothing: float = Field(0.1, frozen=True)
    degrees: float = Field(0.0, frozen=True)
    translate: float = Field(0.1, frozen=True)
    scale: float = Field(0.5, frozen=True)
    flipud: float = Field(0.0, frozen=True)
    fliplr: float = Field(0.5, frozen=True)
    mosaic: float = Field(1.0, frozen=True)
    
    def __init__(self, **data):
        super().__init__(**data)

    def model_post_init(self, __context: Any) -> None:
        super().model_post_init(__context)
        
        self._device_post_init()
        
    def _device_post_init(self):
        if isinstance(self.device, str):
            object.__setattr__(self, 'device', string_to_list_of_type(self.device, int))




import numpy as np
import copy
import os.path as osp
from typing import Union
import matplotlib.pyplot as plt
from shapely.geometry import Polygon
from shapely.affinity import rotate, translate
from scipy.spatial import ConvexHull

def polygon2numpy(polygon_coords: Polygon):
    
    return np.array(list(polygon_coords.exterior.coords)[:-1])

def coords2polygon(coords: Union[list, np.ndarray]):
    polygon_coords = copy.deepcopy(coords)
    if isinstance(polygon_coords, list):
        polygon_coords = np.array(polygon_coords)
    
    if isinstance(polygon_coords, np.ndarray):
        polygon_coords = Polygon(polygon_coords)

    assert isinstance(polygon_coords, Polygon), ValueError(f"The type of input coords must be Polygon, not {type(polygon_coords)}")

    return polygon_coords

def coords2numpy(coords: Union[list, Polygon]):
    numpy_coords = copy.deepcopy(coords)
    if isinstance(numpy_coords, list):
        numpy_coords = np.array(numpy_coords)
    
    if isinstance(numpy_coords, Polygon):
        numpy_coords = polygon2numpy(numpy_coords)

    assert isinstance(numpy_coords, np.ndarray), ValueError(f"The type of input coords must be Numpy, not {type(numpy_coords)}")

    return numpy_coords

def get_obb_coord_by_convex(coords: np.ndarray, flatten:bool = False):
    hull = ConvexHull(coords)
    hull_polygon = Polygon(coords[hull.vertices])

    obb_polygon = hull_polygon.minimum_rotated_rectangle
    obb_coords = np.array(list(obb_polygon.exterior.coords)[:-1])

    if flatten:
        obb_coords = obb_coords.flatten().tolist()
        
    return obb_coords, obb_polygon  

def get_obb_coord_by_roatate(polygon: Polygon, rotate_degree:int = 1, flatten:bool = False):
    def get_obb(polygon, rotate_degree):
        min_area = float('inf')
        best_obb = None
        for angle in range(0, 180, rotate_degree):
            rotated_polygon = rotate(polygon, angle, origin='centroid', use_radians=False)
            minx, miny, maxx, maxy = rotated_polygon.bounds
            area = (maxx - minx) * (maxy - miny)
            if area < min_area:
                min_area = area
                best_obb = (minx, miny, maxx, maxy, angle)
        return best_obb

    minx, miny, maxx, maxy, angle = get_obb(polygon, rotate_degree)
    obb_polygon = Polygon([
        (minx, miny),
        (maxx, miny),
        (maxx, maxy),
        (minx, maxy)
    ])

    obb_polygon = rotate(obb_polygon, -angle, origin='centroid', use_radians=False)
    obb_polygon = translate(obb_polygon, xoff=polygon.centroid.x - obb_polygon.centroid.x, yoff=polygon.centroid.y - obb_polygon.centroid.y)
    obb_coords = polygon2numpy(obb_polygon)
    if flatten:
        obb_coords = obb_coords.flatten().tolist()
    
    return obb_coords, obb_polygon

def xyxy2xywh(imgsz, xyxy):

    if not len(imgsz) >= 2:
        raise RuntimeError(f"imgsz should be [height, width, channel] or [height, width]")
    elif imgsz[0] <= 3:
        raise RuntimeError(f"imgsz should be [height, width, channel] or [height, width]")

    if isinstance(xyxy, list):
        x1, y1, x2, y2 = xyxy[0], xyxy[1], xyxy[2], xyxy[3]
    else:
        raise RuntimeError(f"xyxy must be list such as [x1, y1, x2, y2]")

    def sorting(l1, l2):
        if l1 > l2:
            lmax, lmin = l1, l2
            return lmax, lmin
        else:
            lmax, lmin = l2, l1
            return lmax, lmin

    xmax, xmin = sorting(x1, x2)
    ymax, ymin = sorting(y1, y2)
    
    dw = 1./imgsz[1]
    dh = 1./imgsz[0]
    
    x = (xmin + xmax)/2.0
    y = (ymin + ymax)/2.0
    w = xmax - xmin
    h = ymax - ymin
    
    x = x*dw
    w = w*dw
    y = y*dh
    h = h*dh
    
    return [x, y, w, h]

def xywh2xyxy(imgsz, xywh):

    if not len(imgsz) >= 2:
        raise RuntimeError(f"imgsz should be [height, width, channel] or [height, width]")
    elif imgsz[0] <= 3:
        raise RuntimeError(f"imgsz should be [height, width, channel] or [height, width]")

    if isinstance(xywh, list):
        x, y, w, h = xywh[0], xywh[1], xywh[2], xywh[3]
    else:
        raise RuntimeError(f"xywh must be list such as [x, y, w, h]")

    dh, dw = imgsz[0], imgsz[1]
    l = ((x - w / 2) * dw) # x0
    r = ((x + w / 2) * dw) # x1
    t = ((y - h / 2) * dh) # y0
    b = ((y + h / 2) * dh) # y1
    
    if l < 0:
        l = 0
    if r > dw - 1:
        r = dw - 1
    if t < 0:
        t = 0
    if b > dh - 1:
        b = dh - 1   

    return [l, t, r, b] # x0, y0, x1, y1


from enum import Enum
from typing import Any, Literal, Union

from omegaconf import ListConfig
from pydantic import BaseModel, Field

from aivcommon.configs.utils.helpers import string_to_list_of_type


class DeviceIdsConfig(str, Enum):
    gpu_0 = "0"
    gpu_1 = "1"
    gpu_2 = "2"
    gpu_3 = "3"


class DeviceConfig(BaseModel):

    device: Literal['cpu', 'gpu', 'cuda'] = Field('gpu', frozen=True)
    device_ids: DeviceIdsConfig = Field('0', frozen=True)

    tmp_device_ids: Union[str, None] = Field(None, exclude=True)

    def __init__(self, **data):
        if 'device_ids' in data.keys():
            if isinstance(data['device_ids'], str):
                self._set_tmp_device_ids(data)
            elif isinstance(data['device_ids'], (list, ListConfig)):
                if len(data['device_ids']) == 0:
                    ValueError(f"Device-id must be defined, not f{data['device_ids']}")
                else:
                    if isinstance(data['device_ids'][0], str):
                        data['device_ids'] = ','.join(data['device_ids'])
                        self._set_tmp_device_ids(data)
                    elif isinstance(data['device_ids'][0], int):
                        data['device_ids'] = ','.join(map(str, data['device_ids']))
                        self._set_tmp_device_ids(data)
                    else:
                        NotImplementedError(
                            f"This case is not considered: {type(data['device_ids']), data['device_ids']}")
            elif isinstance(data['device_ids'], int):
                data['device_ids'] = str(data['device_ids'])
            else:
                NotImplementedError(f"This case is not considered: {type(data['device_ids']), data['device_ids']}")

        super().__init__(**data)

    def model_post_init(self, __context: Any) -> None:
        super().model_post_init(__context)
        self._device_post_init()
        self._device_ids_post_init()

    def _set_tmp_device_ids(self, data):
        if len(data['device_ids']) != 0:
            data['tmp_device_ids'] = data['device_ids']
            data['device_ids'] = data['device_ids'][0]
        else:
            ValueError(f"Device-id must be defined, not f{data['device_ids']}")

    def _device_ids_post_init(self):
        if self.tmp_device_ids is not None:
            object.__setattr__(self, 'device_ids', string_to_list_of_type(self.tmp_device_ids, int))

        elif isinstance(self.device_ids, str):
            object.__setattr__(self, 'device_ids', string_to_list_of_type(self.device_ids, int))

    def _device_post_init(self):
        if self.device == 'gpu':
            object.__setattr__(self, 'device', 'cuda')

    @property
    def device_ids_list(self):
        if isinstance(self.device_ids, str):
            return string_to_list_of_type(self.device_ids, int)
        else:
            return self.device_ids



from pathlib import Path
import cv2 
import os
import os.path as osp
from tqdm import tqdm
from shutil import copyfile
VERBOSE = str(os.getenv("YOLO_VERBOSE", True)).lower() == "true"  # global verbose mode
TQDM_BAR_FORMAT = "{l_bar}{bar:10}{r_bar}" if VERBOSE else None  # tqdm bar format


def convert_dota2yolo_obb(dota_root_path: str, save_dir: str, 
                          image_width: int=None, image_height: int=None,
                          class2idx: dict={}, 
                          copy_image: bool=True, image_ext='bmp'):

    dota_root_path = Path(dota_root_path)
    def convert_label(image_name, image_width, image_height, orig_label_dir, save_dir):
        """Converts a single image's DOTA annotation to YOLO OBB format and saves it to a specified directory."""
        orig_label_path = orig_label_dir / f"{image_name}.txt"
        save_path = save_dir / f"{image_name}.txt"

        with orig_label_path.open("r") as f, save_path.open("w") as g:
            lines = f.readlines()
            for line in lines:
                parts = line.strip().split()
                if len(parts) < 9:
                    continue
                class_name = parts[8]
                if class_name in class2idx:
                    class_idx = class2idx[class_name]
                else:
                    class_idx = len(class2idx)
                    class2idx[class_name] = class_idx
                coords = [float(p) for p in parts[:8]]
                normalized_coords = [
                    coords[i] / image_width if i % 2 == 0 else coords[i] / image_height for i in range(8)
                ]
                formatted_coords = ["{:.6g}".format(coord) for coord in normalized_coords]
                g.write(f"{class_idx} {' '.join(formatted_coords)}\n")

    for phase in ["train", "val"]:
        if copy_image:
            image_dir = dota_root_path  / phase/ "images"
            orig_label_dir = dota_root_path / phase / "labelTxt" 
            save_labels_dir = Path(save_dir) / "labels" / phase
            save_labels_dir.mkdir(parents=True, exist_ok=True)
            save_images_dir = Path(save_dir) / "images" / phase
            save_images_dir.mkdir(parents=True, exist_ok=True)
            image_paths = list(image_dir.iterdir())
            for image_path in tqdm(image_paths, desc=phase):
                image_name_without_ext = image_path.stem
                copyfile(image_path, save_images_dir / Path(image_path.stem + image_path.suffix))
                img = cv2.imread(str(image_path))
                h, w = img.shape[:2]
                convert_label(image_name_without_ext, w, h, orig_label_dir, save_labels_dir)
        else:
            assert image_width is not None and image_height is not None, ValueError(f"Height({image_height}) and Width({image_width}) of image must not be None")
            orig_label_dir = dota_root_path / phase / "labelTxt" 
            save_labels_dir = Path(save_dir) / "labels" / phase
            save_labels_dir.mkdir(parents=True, exist_ok=True)
            label_paths = list(orig_label_dir.iterdir())
            for label_path in tqdm(label_paths, desc=phase):
                image_name_without_ext = label_path.stem
                h, w = img.shape[:2]
                convert_label(image_name_without_ext, image_width, image_height, orig_label_dir, save_labels_dir)


    idx2class_txt = open(osp.join(save_dir, 'idx2class.txt'), 'w')
    class2idx_txt = open(osp.join(save_dir, 'class2idx.txt'), 'w')
    for key, val in class2idx.items():
        idx2class_txt.write(f"{val}: {key}\n")
        class2idx_txt.write(f"{key}: {val}\n")
        
    idx2class_txt.close()
    class2idx_txt.close()



import numpy as np
import os.path as osp
from typing import Union
import matplotlib.pyplot as plt
from shapely.geometry import Polygon
from mlearning.datasets.converters.utils import get_obb_coord_by_roatate, get_obb_coord_by_convex, coords2polygon, coords2numpy

def polygon2dota_by_rotate(coords: Union[list, np.ndarray, Polygon], rotate_degree:int = 1, flatten: bool = False, output_dir: str = None):
    polygon = coords2polygon(coords)
    obb_coords, obb_polygon = get_obb_coord_by_roatate(polygon, rotate_degree, flatten)

    if output_dir and osp.exists(output_dir):
        plt.figure(figsize=(10, 10))
        plt.plot(*polygon.exterior.xy, color='blue', label='Original Polygon')
        plt.plot(*obb_polygon.exterior.xy, color='red', linestyle='--', label='Oriented Bounding Box')
        # plt.scatter(coords[:, 0], coords[:, 1], color='blue')
        plt.scatter(obb_coords[:, 0], obb_coords[:, 1], color='red')
        plt.legend()
        plt.xlabel('X')
        plt.ylabel('Y')
        plt.title('Polygon and its Oriented Bounding Box')
        plt.show()
        plt.savefig(osp.join(output_dir, "obb.png"))
        
    return obb_coords, obb_polygon

def polygon2dota_by_convex(coords: Union[list, np.ndarray, Polygon], flatten: bool = False, output_dir: str = None):
    numpy_coords = coords2numpy(coords)
    obb_coords, obb_polygon = get_obb_coord_by_convex(numpy_coords, flatten)

    if output_dir and osp.exists(output_dir):
        plt.figure(figsize=(10, 10))
        closed_numpy_coords = np.vstack([numpy_coords, numpy_coords[0]])
        plt.plot(closed_numpy_coords[:, 0], closed_numpy_coords[:, 1], color='blue', label='Original Polygon')
        plt.plot(*obb_polygon.exterior.xy, color='red', linestyle='--', label='Oriented Bounding Box')
        plt.scatter(numpy_coords[:, 0], numpy_coords[:, 1], color='blue')
        plt.scatter(obb_coords[:, 0], obb_coords[:, 1], color='red')
        plt.legend()
        plt.xlabel('X')
        plt.ylabel('Y')
        plt.title('Polygon and its Oriented Bounding Box')
        plt.show()
        plt.savefig(osp.join(output_dir, "obb.png"))

    return obb_coords, obb_polygon

if __name__ == '__main__':
    polygon_coords = [
        [1960.2086065298297, 788.6438090123423],
        [2332.7453506990687, 871.5613673329996],
        [2332.7453506990687, 871.5613673329996],
        [2227.913866250809, 1355.4445469614068],
        [1861.2998048187603, 1276.6728665567823]
    ]
    
    # polygon2dota_by_rotate(Polygon(np.array(polygon_coords)), 1, output_dir='/HDD/_projects/nothing')
    polygon2dota_by_convex(Polygon(np.array(polygon_coords)), output_dir='/HDD/_projects/nothing')